
#include "../elements.h"
#include "minxp/array.h"
#include <exception>
#include <iostream>
#include <cstdio>

void exit_usage()
{
	std::cerr << "chksort -e{i|p|r} file [source]" << std::endl;
	exit(1);
}

template<class T>
void check_sort(const char *file, const char *source)
{
	T chksum;
	typename minxp::map_array<T> A(file, 0, minxp::MAP_EXISTING | minxp::MAP_READ | minxp::MAP_WRITE);
	typename minxp::map_array<T>::iterator p = A.begin();
	chksum = *p;
	T last = *p;
	for( ++p; p!=A.end(); ++p )
	{
		if( *p < last )
			std::cout << "Elements " << p-A.begin()-1 << " and " << p-A.begin() << " are out of order: " << last << '>' << '=' << *p << '!' << std::endl;
		chksum += *p;
		last = *p;
	}
	A.free();
	if( source )
	{
		typename minxp::map_array<T> A(source, 0, minxp::MAP_EXISTING | minxp::MAP_READ | minxp::MAP_WRITE);
		typename minxp::map_array<T>::iterator p = A.begin();
		T outsum = *p;
		for( ++p; p!=A.end(); ++p )
			outsum += *p;
		if(!( outsum == chksum ))
			std::cout << "Checksum error: " << chksum << " != " << outsum << '!' << std::endl;
	}
}

int main(int argc, char *argv[])
{
	if( argc < 3 || argc > 4 )
		exit_usage();
	if( argv[1][0] != '-' || argv[1][1] != 'e' )
		exit_usage();
	try
	{
		switch( argv[1][2] )
		{
		case 'i':
			check_sort<int>(argv[2], argc==4? argv[3] : NULL);
			break;
		case 'p':
			check_sort<element>(argv[2], argc==4? argv[3] : NULL);
			break;
		case 'r':
			check_sort<record>(argv[2], argc==4? argv[3] : NULL);
			break;
		default:
			exit_usage();
		}
	}
	catch( std::exception& e )
	{
		std::cerr << "Exception in sortgen: " << e.what() << std::endl;
		return 1;
	}
	return 0;
}
