#ifndef RANDOM_H_INCLUDED__
#define RANDOM_H_INCLUDED__

#include <ctime>
#include <cstdlib>
#include <limits>

class random_source
{
public:
#ifndef _DEBUG
	inline random_source()
	{
#ifdef WIN32
		srand(static_cast<unsigned>(time(NULL)));
#else // WIN32
		srand48(static_cast<long int>(time(NULL)));
#endif // WIN32
	}
#endif// _DEBUG
};
template<class T>
inline random_source& operator>>(random_source& s, T& t);
#ifdef WIN32
template<>
inline random_source& operator>>(random_source& s, int& t)
{
	t = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
	return s;
}
template<>
inline random_source& operator>>(random_source& s, double& t)
{
	int i;
	s >> i;
	t = static_cast<double>(i)/(static_cast<double>(std::numeric_limits<int>::max())-std::numeric_limits<int>::min())+0.5;
	return s;
}
#else // WIN32
template<>
inline random_source& operator>>(random_source& s, double& t)
{
	t = drand48();
	return s;
}
template<>
inline random_source& operator>>(random_source& s, int& t)
{
	t = mrand48();
	return s;
}
template<>
inline random_source& operator>>(random_source& s, unsigned int& t)
{
	t = static_cast<unsigned int>(lrand48());
	return s;
}
#endif // WIN32
template<class T>
inline random_source& operator>>(random_source& s, T& t)
{
	double d;
	s >> d;
	t = factory<T>::from_scalar(d);
	return s;
}

template<class T>
class random_stream
{
public:
	class iterator : public std::iterator<std::forward_iterator_tag,T>
	{
		friend class random_stream<T>;
	public:
		inline iterator& operator++()
			{ ++N; return *this; }
		inline T operator*()
			{ T t; s >> t; return t; }
		inline bool operator==(const iterator& rhs)
			{ return N == rhs.N; }
	private:
		iterator(random_source& s, size_t N) : s(s), N(N) { }
		random_source& s;
		size_t N;
	};
	inline random_stream(size_t N) : N(N) { }
	inline iterator begin()
		{ return iterator(s,0); }
	inline iterator end()
		{ return iterator(s,N); }
private:
	random_source s;
	size_t N;
};

#endif // RANDOM_H_INCLUDED__
