#ifndef __BENCH_H_INCLUDED_
#define __BENCH_H_INCLUDED_

#include "minxp/array.h"
#include "minxp/stopwatch.h"
#include "../elements.h"
#include "../SortGen/sortgen.h"
#include <exception>
#include <iostream>
#include <algorithm>
#include <sstream>

struct measure
{
	measure() : time(), pf(), l1(), l2(), l3(), tlb(), bmsp(), ins() { }
	measure(const minxp::stopwatch& w, const minxp::pf_counter& pfc, const minxp::miss_counter& mc) : time(w.time()), pf(pfc.count()), l1(mc.l1()), l2(mc.l2()), l3(mc.l3()), tlb(mc.tlb()), bmsp(mc.bmsp()), ins(mc.ins()) { }
	bool operator<(const measure& rhs) const
		{ return time < rhs.time; }

	void min(const measure& m)
	{
		cmov(time,m.time,std::greater<double>());
		cmov(pf,m.pf,std::greater<unsigned long>());
		cmov(l1,m.l1,std::greater<minxp::uint64_t>());
		cmov(l2,m.l2,std::greater<minxp::uint64_t>());
		cmov(l3,m.l3,std::greater<minxp::uint64_t>());
		cmov(tlb,m.tlb,std::greater<minxp::uint64_t>());
		cmov(bmsp,m.bmsp,std::greater<minxp::uint64_t>());
		cmov(ins,m.ins,std::greater<minxp::uint64_t>());
	}
	void max(const measure& m)
	{
		cmov(time,m.time,std::less<double>());
		cmov(pf,m.pf,std::less<unsigned long>());
		cmov(l1,m.l1,std::less<minxp::uint64_t>());
		cmov(l2,m.l2,std::less<minxp::uint64_t>());
		cmov(l3,m.l3,std::less<minxp::uint64_t>());
		cmov(tlb,m.tlb,std::less<minxp::uint64_t>());
		cmov(bmsp,m.bmsp,std::less<minxp::uint64_t>());
		cmov(ins,m.ins,std::less<minxp::uint64_t>());
	}

	double time;
	unsigned long pf;
	minxp::uint64_t l1, l2, l3, tlb, bmsp, ins;
private:
	template<class T, class Pred>
	static inline void cmov(T& a, const T& b, Pred p)
	{
		if( p(a,b) )
			a = b;
	}
};
std::ostream& operator<<(std::ostream& s, const measure& m)
{
	return s << m.time;
}
namespace std
{
template<>
class numeric_limits<measure>
{
public:
	inline static measure min()
	{
		measure ret;
		ret.time = std::numeric_limits<double>::min();
		ret.pf = std::numeric_limits<unsigned long>::min();
		ret.l1 = ret.l2 = ret.l3 = ret.tlb = ret.bmsp = ret.ins = std::numeric_limits<minxp::uint64_t>::min();
		return ret;
	}
	inline static measure max()
	{
		measure ret;
		ret.time = std::numeric_limits<double>::max();
		ret.pf = std::numeric_limits<unsigned long>::max();
		ret.l1 = ret.l2 = ret.l3 = ret.tlb = ret.bmsp = ret.ins = std::numeric_limits<minxp::uint64_t>::max();
		return ret;
	}
};
} // namespace std

volatile int res;
int *A = NULL;

void fill_cache()
{
	const int size = 1024*1024;
	A = new int[size];
	A[0] = rand();
	A[1] = rand();
	for( int i=2; i!=size; ++i )
		A[i] = 2*A[i-2]-A[i-1];
	res = A[size-1] - A[rand() % size];
	delete[] A;
}


template<class T>
struct type_stat { };
template<class T>
inline std::ostream& operator<<(std::ostream& s, type_stat<T>)
{
	return s;
}

template<class T>
struct type_stat<minxp::counting<T> > { };
template<class T>
inline std::ostream& operator<<(std::ostream& s, type_stat<minxp::counting<T> >)
{
	return s << "<comparisons>" << minxp::counting<T>::comparisons() << "</comparisons><moves>" << minxp::counting<T>::moves() << "</moves>";
}

template<class Sorter, class T>
void do_int_sort(size_t size, char dist, minxp::false_type)
{
	minxp::stopwatch w;
	minxp::pf_counter pfc;
	minxp::miss_counter mc;
	typename minxp::dyn_array<T> B(size);
	std::cout << "\t\t<size>" << size << "</size>" << std::endl;
	std::vector<measure> samples;
	measure m, fastest = std::numeric_limits<measure>::max(), slowest = std::numeric_limits<measure>::min();
	for( int j=0; j<12; ++j )
	{
		typename minxp::dyn_array<T> A(size);
		for( int i=0; i<21; ++i )
		{
			sortgen<T>(A.begin(),A.end(),dist);
			w.reset(); mc.reset(); pfc.reset();
			fill_cache();
			pfc.start(); mc.start(); w.start();
			Sorter::sort(A.begin(), A.end(), B.begin());
			w.stop(); mc.stop(); pfc.stop();
			m = measure(w,pfc,mc);
			samples.push_back(m);
			fastest.min(m);
			slowest.max(m);
		}
		std::sort(samples.begin(), samples.end());
		if( samples[0].time > 0.8*samples[samples.size()/2].time && samples[samples.size()-1].time < 1.2*samples[samples.size()/2].time )
			break;
		samples[0] = samples[samples.size()-2];
		samples.pop_back(); samples.pop_back();
		A.free();
	}
	B.free();
	m = samples[samples.size()/2];
	std::cout << "\t\t<time unit=\"s\">" << m.time << "</time><time-err unit=\"s\" dir=\"plus\">" << slowest.time-m.time << "</time-err><time-err unit=\"s\" dir=\"minus\">" << m.time-fastest.time << "</time-err>\n"
#ifdef PAPI
		<< "\t\t<l1misses>" << m.l1 << "</l1misses><l1misses-err dir=\"plus\">" << slowest.l1-m.l1 << "</l1misses-err><l1misses-err dir=\"minus\">" << m.l1-fastest.l1 << "</l1misses-err>\n"
		<< "\t\t<l2misses>" << m.l2 << "</l2misses><l2misses-err dir=\"plus\">" << slowest.l2-m.l2 << "</l2misses-err><l2misses-err dir=\"minus\">" << m.l2-fastest.l2 << "</l2misses-err>\n"
		<< "\t\t<l3misses>" << m.l3 << "</l3misses><l3misses-err dir=\"plus\">" << slowest.l3-m.l3 << "</l3misses-err><l3misses-err dir=\"minus\">" << m.l3-fastest.l3 << "</l3misses-err>\n"
		<< "\t\t<tlbmisses>" << m.tlb << "</tlbmisses><tlbmisses-err dir=\"plus\">" << slowest.tlb-m.tlb << "</tlbmisses-err><tlbmisses-err dir=\"minus\">" << m.tlb-fastest.tlb << "</tlbmisses-err>\n"
		<< "\t\t<branch-misspredict>" << m.bmsp << "</branch-misspredict><branch-misspredict-err dir=\"plus\">" << slowest.bmsp-m.bmsp << "</branch-misspredict-err><branch-misspredict-err dir=\"minus\">" << m.bmsp-fastest.bmsp << "</branch-misspredict-err>\n"
		<< "\t\t<instruction-count>" << m.ins << "</instruction-count><instruction-count-err dir=\"plus\">" << slowest.ins-m.ins << "</instruction-count-err><instruction-count-err dir=\"minus\">" << m.ins-fastest.ins << "</instruction-count-err>\n"
#endif // PAPI
		<< "\t\t<pagefaults>" << m.pf << "</pagefaults><pagefaults-err dir=\"plus\">" << slowest.pf-m.pf << "</pagefaults-err><pagefaults-err dir=\"minus\">" << m.pf-fastest.pf << "</pagefaults-err>" << std::endl
		<< '\t' << '\t' << type_stat<T>() << std::endl;
}

template<class Sorter, class T>
void do_single_sort(size_t size, char dist, minxp::false_type)
{
	minxp::stopwatch w;
	minxp::pf_counter pfc;
	minxp::miss_counter mc;
	typename minxp::dyn_array<T> B(size);
	typename minxp::dyn_array<T> A(size);
	sortgen<T>(A.begin(),A.end(),dist);
	std::cout << "\t\t<size>" << size << "</size>" << std::endl;
	w.reset(); mc.reset(); pfc.reset();
	fill_cache();
	pfc.start(); mc.start(); w.start();
	Sorter::sort(A.begin(), A.end(), B.begin());
	w.stop(); mc.stop(); pfc.stop();
	A.free();
	B.free();
	std::cout << w.xmltag(2) << std::endl
		<< pfc.xmltag(2) << std::endl
		<< mc.xmltag(2) << std::endl
		<< '\t' << '\t' << type_stat<T>() << std::endl;
}

template<class Sorter, class T>
void do_ext_sort(const char *in, const char *out, std::output_iterator_tag)
{
	typename minxp::map_array<T> A(in, 0, minxp::MAP_EXISTING | minxp::MAP_READ | minxp::MAP_WRITE);
	minxp::stopwatch w;
	minxp::pf_counter pfc;
	minxp::miss_counter mc;
	std::cout << "\t\t<size>" << static_cast<long>(A.end()-A.begin()) << "</size>" << std::endl;
	typename minxp::binfile_ostream<T> B(out);
	pfc.start(); mc.start(); w.start();
	Sorter::sort(A.begin(), A.end(), B.begin());
	w.stop(); mc.stop(); pfc.stop();
	A.free();
	std::cout << w.xmltag(2) << std::endl
		<< pfc.xmltag(2) << std::endl
		<< mc.xmltag(2) << std::endl
		<< '\t' << '\t' << type_stat<T>() << std::endl;
}

template<class Sorter, class T>
void do_ext_sort(const char *in, const char *out, std::forward_iterator_tag)
{
	typename minxp::map_array<T> A(in, 0, minxp::MAP_EXISTING | minxp::MAP_READ | minxp::MAP_WRITE);
	minxp::stopwatch w;
	minxp::pf_counter pfc;
	minxp::miss_counter mc;
	std::cout << "\t\t<size>" << static_cast<long>(A.end()-A.begin()) << "</size>" << std::endl;
	typename minxp::map_array<T> B(out, A.end()-A.begin(), minxp::MAP_CREATE | minxp::MAP_READ | minxp::MAP_WRITE);
	pfc.start(); mc.start(); w.start();
	Sorter::sort(A.begin(), A.end(), B.begin());
	w.stop(); mc.stop(); pfc.stop();
	B.free();
	A.free();
	std::cout << w.xmltag(2) << std::endl
		<< pfc.xmltag(2) << std::endl
		<< mc.xmltag(2) << std::endl
		<< '\t' << '\t' << type_stat<T>() << std::endl;
}

template<class Sorter, class T>
void do_ext_sort(const char *in, const char *out, minxp::false_type)
{
	do_ext_sort<Sorter,T>(in, out, typename Sorter::out_iterator_category());
}

template<class Sorter, class T>
void do_int_sort(size_t size, char dist, minxp::true_type)
{
	minxp::stopwatch w;
	minxp::pf_counter pfc;
	minxp::miss_counter mc;
	std::cout << "\t\t<size>" << size << "</size>" << std::endl;
	std::vector<measure> samples;
	measure m, fastest = std::numeric_limits<measure>::max(), slowest = std::numeric_limits<measure>::min();
	for( int j=0; j<12; ++j )
	{
		typename minxp::dyn_array<T> A(size);
		for( int i=0; i<21; ++i )
		{
			sortgen<T>(A.begin(),A.end(),dist);
			w.reset(); mc.reset(); pfc.reset();
			fill_cache();
			pfc.start(); mc.start(); w.start();
			Sorter::sort(A.begin(), A.end());
			w.stop(); mc.stop(); pfc.stop();
			m = measure(w,pfc,mc);
			samples.push_back(m);
			fastest.min(m);
			slowest.max(m);
		}
		std::sort(samples.begin(), samples.end());
		if( samples[0].time > 0.8*samples[samples.size()/2].time && samples[samples.size()-1].time < 1.2*samples[samples.size()/2].time )
			break;
		samples[0] = samples[samples.size()-2];
		samples.pop_back(); samples.pop_back();
		A.free();
	}
	m = samples[samples.size()/2];
	std::cout << "\t\t<time unit=\"s\">" << m.time << "</time><time-err unit=\"s\" dir=\"plus\">" << slowest.time-m.time << "</time-err><time-err unit=\"s\" dir=\"minus\">" << m.time-fastest.time << "</time-err>\n"
#ifdef PAPI
		<< "\t\t<l1misses>" << m.l1 << "</l1misses><l1misses-err dir=\"plus\">" << slowest.l1-m.l1 << "</l1misses-err><l1misses-err dir=\"minus\">" << m.l1-fastest.l1 << "</l1misses-err>\n"
		<< "\t\t<l2misses>" << m.l2 << "</l2misses><l2misses-err dir=\"plus\">" << slowest.l2-m.l2 << "</l2misses-err><l2misses-err dir=\"minus\">" << m.l2-fastest.l2 << "</l2misses-err>\n"
		<< "\t\t<l3misses>" << m.l3 << "</l3misses><l3misses-err dir=\"plus\">" << slowest.l3-m.l3 << "</l3misses-err><l3misses-err dir=\"minus\">" << m.l3-fastest.l3 << "</l3misses-err>\n"
		<< "\t\t<tlbmisses>" << m.tlb << "</tlbmisses><tlbmisses-err dir=\"plus\">" << slowest.tlb-m.tlb << "</tlbmisses-err><tlbmisses-err dir=\"minus\">" << m.tlb-fastest.tlb << "</tlbmisses-err>\n"
		<< "\t\t<branch-misspredict>" << m.bmsp << "</branch-misspredict><branch-misspredict-err dir=\"plus\">" << slowest.bmsp-m.bmsp << "</branch-misspredict-err><branch-misspredict-err dir=\"minus\">" << m.bmsp-fastest.bmsp << "</branch-misspredict-err>\n"
		<< "\t\t<instruction-count>" << m.ins << "</instruction-count><instruction-count-err dir=\"plus\">" << slowest.ins-m.ins << "</instruction-count-err><instruction-count-err dir=\"minus\">" << m.ins-fastest.ins << "</instruction-count-err>\n"
#endif // PAPI
		<< "\t\t<pagefaults>" << m.pf << "</pagefaults><pagefaults-err dir=\"plus\">" << slowest.pf-m.pf << "</pagefaults-err><pagefaults-err dir=\"minus\">" << m.pf-fastest.pf << "</pagefaults-err>" << std::endl
		<< '\t' << '\t' << type_stat<T>() << std::endl;
}

template<class Sorter, class T>
void do_single_sort(size_t size, char dist, minxp::true_type)
{
	minxp::stopwatch w;
	minxp::pf_counter pfc;
	minxp::miss_counter mc;
	typename minxp::dyn_array<T> A(size);
	sortgen<T>(A.begin(),A.end(),dist);
	std::cout << "\t\t<size>" << size << "</size>" << std::endl;
	w.reset(); mc.reset(); pfc.reset();
	fill_cache();
	pfc.start(); mc.start(); w.start();
	Sorter::sort(A.begin(), A.end());
	w.stop(); mc.stop(); pfc.stop();
	A.free();
	std::cout << w.xmltag(2) << std::endl
		<< pfc.xmltag(2) << std::endl
		<< mc.xmltag(2) << std::endl
		<< '\t' << '\t' << type_stat<T>() << std::endl;
}

template<class Sorter, class T>
void do_ext_sort(const char *in, const char *out, minxp::true_type)
{
	typename minxp::map_array<T> A(in, 0, minxp::MAP_EXISTING | minxp::MAP_READ | minxp::MAP_WRITE);
	minxp::stopwatch w;
	minxp::pf_counter pfc;
	minxp::miss_counter mc;
	std::cout << "\t\t<size>" << static_cast<long>(A.end()-A.begin()) << "</size>" << std::endl;
	pfc.start(); mc.start(); w.start();
	Sorter::sort(A.begin(), A.end());
	w.stop(); mc.stop(); pfc.stop();
	A.free();
	std::cout << w.xmltag(2) << std::endl
		<< pfc.xmltag(2) << std::endl
		<< mc.xmltag(2) << std::endl
		<< '\t' << '\t' << type_stat<T>() << std::endl;
}


struct usage { };

template<class sorter>
int bench_this(char **argb, char **arge)
{
	try
	{
		if( arge-argb != 4 )
			throw usage();
		if( argb[1][0] != '-' || argb[1][1] != 'e' )
			throw usage();
		if( argb[2][0] == '-' )
		{
			if( argb[2][1] != 'd' )
				throw usage();
			std::istringstream ss(argb[3]);
			size_t size = 0;
			if( !(ss >> size) )
				throw usage();
			std::cout << "\t<run buildtime=\"" << minxp::sys_names::buildtime() << "\" compiler=\"" << minxp::sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
			std::cout << "\t\t<algorithm>" << sorter::name << "</algorithm>" << std::endl;
			std::cout << "\t\t<distribution>";
			switch( argb[2][2] )
			{
			case 'u':
				std::cout << "uniform";
				break;
			case 'k':
				std::cout << "killer";
				break;
			case 'i':
				std::cout << "identical";
				break;
			case 'm':
				std::cout << "missing-merge(log)";
				break;
			case 'M':
				std::cout << "Missing-Merge(sqrt)";
				break;
			case 'd':
				std::cout << "missing-distribution(log)";
				break;
			case 'D':
				std::cout << "Missing-Distribution(sqrt)";
				break;
			case 'e':
				std::cout << "few-disitinct(log)";
				break;
			case 'E':
				std::cout << "few-disitinct(sqrt)";
				break;
			case 'r':
				std::cout << "few-swaps(log)";
				break;
			case 'R':
				std::cout << "few-swaps(sqrt)";
				break;
			default:
				throw usage();
			}
			std::cout << "</distribution>" << std::endl;
			switch( argb[1][2] )
			{
				case 'i':
					std::cout << "\t\t<type>int</type>" << std::endl;
					do_int_sort<sorter,int>(size,argb[2][2],typename sorter::inplace());
					break;
				case 'p':
					std::cout << "\t\t<type>pair</type>" << std::endl;
					do_int_sort<sorter,element>(size,argb[2][2],typename sorter::inplace());
					break;
				case 'r':
					std::cout << "\t\t<type>record</type>" << std::endl;
					do_int_sort<sorter,record>(size,argb[2][2],typename sorter::inplace());
					break;
				case 'c':
					std::cout << "\t\t<type>counting long</type>" << std::endl;
					do_single_sort<sorter,minxp::counting<long> >(size,argb[2][2],typename sorter::inplace());
					break;
				break;
			default:
				std::cout << "\t</run>" << std::endl << std::endl;
				throw usage();
			}
		}
		else
		{
			std::cout << "\t<run buildtime=\"" << minxp::sys_names::buildtime() << "\" compiler=\"" << minxp::sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
			std::cout << "\t\t<algorithm>" << sorter::name << "</algorithm>" << std::endl;
			switch( argb[1][2] )
			{
				case 'i':
					std::cout << "\t\t<type>int</type>" << std::endl;
					do_ext_sort<sorter,int>(argb[2],argb[3],typename sorter::inplace());
					break;
				case 'p':
					std::cout << "\t\t<type>pair</type>" << std::endl;
					do_ext_sort<sorter,element>(argb[2],argb[3],typename sorter::inplace());
					break;
				case 'r':
					std::cout << "\t\t<type>counting long</type>" << std::endl;
					do_ext_sort<sorter,minxp::counting<long> >(argb[2],argb[3],typename sorter::inplace());
					break;
				break;
			default:
				std::cout << "\t</run>" << std::endl << std::endl;
				throw usage();
			}
		}
	}
	catch( usage& )
	{
		std::cerr << sorter::name << " -e{i|p|r} input output" << std::endl
		          << sorter::name << " -e{i|p|r} -d{u|i|m|M|d|D|s|S|e|t|k} size" << std::endl;
		return 1;
	}
	catch( minxp::exception& ex )
	{
		std::cout << "\t\t<error>" << ex << "</error>\n\t</run>" << std::endl;
		return 1;
	}
	catch( std::exception& ex )
	{
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>\n\t</run>" << std::endl;
		return 1;
	}
	std::cout << "\t</run>" << std::endl;
	return 0;
}

#endif // __BENCH_H_INCLUDED_
