
#include <cassert>
#include "minxp/array.h"
#include "../bench.h"
#include "FunnelSortOld/funnel.h"
#include "FunnelSortOld/sort.h"
#include <memory>
#include <iterator>

template<class T>
class sorted_stream
{
public:
	class iterator;
	friend class iterator;
	class iterator : public std::iterator<std::output_iterator_tag, T, int>
	{
		friend class sorted_stream<T>;
	public:
		inline iterator& operator++()
		{
			if( s->cnt )
			{
				if( s->current < s->prev )
					std::cerr << "Elements " << s->cnt << " and " << s->cnt+1 << " out of order: " << s->prev << '>' << s->current << std::endl;
				s->prev = s->current;
				s->chksum += s->current;
			}
			else
			{
				s->prev = s->current;
				s->chksum = s->current;
			}
			++s->cnt;
			return *this;
		}
		inline T& operator*()
			{ return s->current; }
		inline const T& operator*() const
			{ return s->current; }
	private:
		iterator(sorted_stream *s) : s(s) { }
		sorted_stream *s;
	};

	sorted_stream(size_t s) : size(s), cnt(0) { }
	iterator begin()
		{ return iterator(this); }
	inline T checksum() const
		{ return chksum; }
	inline size_t inserted() const
		{ return cnt; }
protected:
	T prev;
	T current;
	T chksum;
	size_t size;
	size_t cnt;
};
template<class Merger>
void check_heap_merger(bool balanced = false)
{
	using namespace iosort;
	order_t k = static_cast<order_t>((rand() & 63) + Merger::order + 1);
	size_t N = k*k*k/2;
	size_t ss = k*k;
	if( balanced )
		k = pow_of_order<Merger::order>(logc<Merger::order>(k));
	N = N+(size_t)((float)rand()/RAND_MAX*N);
	std::cout << "Checking " << k << "-merger..." << std::flush;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	T *input = new T[N];
	sortgen<T>(input,input+N,'u');
	T *p;
	std::cout << '.' << std::flush;
	std::allocator<T> alloc;
	{
		Merger merge_tree(k,alloc);
		std::cout << '.' << std::flush;
		for( p=input; p+ss<=input+N; p+=ss )
			std::sort(p, p+ss);
		if( p < input+N )
			std::sort(p, input+N);
		std::cout << '.' << std::flush;
		for( int i=0; i<3; ++i )
		{
			sorted_stream<T> output(N);
			for( p=input; p+ss<=input+N; p+=ss)
				merge_tree.add_stream(p,p+ss);
			if( p < input+N )
				merge_tree.add_stream(p,input+N);
			merge_tree(output.begin()/*, output.end()*/);
			assert( output.inserted() == N );
//			assert( output.checksum() == chksum );
			merge_tree.reset();
		}
	}
/*	{
		sorted_stream<T> output(N);
		typename Merger::template rebind<Refill42>::other merge_tree42(k,alloc);
		for( p=input; p+ss<=input+N; p+=ss)
			merge_tree42.add_stream(stream(p,p+ss));
		if( p < input+N )
			merge_tree42.add_stream(stream(p,input+N));
		typename sorted_stream<T>::iterator e = merge_tree42(output.begin(), output.end());
		assert( e == output.end() );
		assert( output.checksum() == chksum );
		for( p=input; p<input+N; ++p )
			assert( *p == 42 );
		std::cout << '.';
	}
*/	std::cout << std::endl;
	delete[] input;
}
void checkFSort()
{
//	typedef minxp::safe_array<element> Array;
	typedef minxp::dyn_array<element> Array;
	for( unsigned int size=29; size<19000; size++ )
	{
		Array::value_type sum = Array::value_type();
		Array src(size);
		Array dest(size);
		for( size_t i=0; i<size; i++ )
		{
//			src[i] = rand();
			src[i] = element(rand(),NULL);
			sum += src[i];
		}
		std::cout << "Sorting " << size << " elements..." << std::flush;
		iosort::merge_sort<iosort::merge_tree<Array::iterator,4>,iosort::default_splitter<4> >(src.begin(), src.end(), dest.begin());
		sorted_stream<element> ss(size);
		std::copy(dest.begin(),dest.end(),ss.begin());
		assert( ss.checksum() == sum );
		std::cout << " and back again..." << std::flush;
		iosort::merge_sort<iosort::merge_tree<Array::iterator,4>,iosort::default_splitter<4> >(dest.begin(), dest.end(), src.begin());
		sorted_stream<element> ssb(size);
		std::copy(src.begin(),src.end(),ssb.begin());
		assert( ssb.checksum() == sum );
		std::cout << " done" << std::endl;
		dest.free();
		src.free();
	}
}

class splitter_two : public iosort::default_splitter<2>
{
public:
	static inline size_t switch_to_cache_aware()
		{ return 16*(4+1)*(4+1); }
};
typedef iosort::default_splitter<4> splitter_four;

struct sorter
{
	typedef minxp::false_type inplace;
	typedef std::output_iterator_tag out_iterator_category;
	template<class RanIt, class OutIt>
	static inline void sort(RanIt b, RanIt e, OutIt o)
	{
		typedef typename std::iterator_traits<RanIt>::value_type T;
		typedef typename iosort::merge_tree<T*,4> merger;
		std::allocator<T> alloc;
		iosort::merge_sort<merger,splitter_four>(b, e, o, alloc);
	}
	static const char* name;
};
const char* sorter::name = "funnelsort4";

int main(int argc, char *argv[])
{
	if( argc == 1 )
	{
		checkFSort();
		for( ;; )
			check_heap_merger<iosort::merge_tree<element*,4> >();
		return 0;
	}
	else
		return bench_this<sorter>(argv,argc+argv);
}
