
#include <algorithm>
#include <functional>
#include <utility>
#include <iterator>
#include <cassert>

enum tag_RECURSELIMIT_ { RECURSELIMIT_ = 24 };

template<class It, class Pred>
inline void insertion_sort(It begin, It end, Pred comp)
{
	for( It j=begin+1; j<end; j++ )
	{
		typename std::iterator_traits<It>::value_type e = *j;
		for( It i=j; begin<i; *i=e )
			if( comp(e,*(i-1)) )
				*i = *(i-1), --i;
			else
				break;
	}
}

template<typename T, class Pred>
inline T median_(const T& a, const T& b, const T& c, Pred comp)
{
	if( comp(a,b) )
		if( comp(b,c) )
			return b;
		else if( comp(a,c) )
			return c;
		else
			return a;
	else
		if( comp(a,c) )
			return a;
		else if( comp(b,c) )
			return c;
		else
			return b;
}

template<class RanIt, class Pred>
inline void med3_(RanIt First, RanIt Mid, RanIt Last, Pred comp)
{
	if( comp(*Mid,*First) )
		std::iter_swap(Mid, First);
	if( comp(*Last,*Mid) )
		std::iter_swap(Last, Mid);
	if( comp(*Mid,*First) )
		std::iter_swap(Mid, First);
}

template<class RanIt, class Pred>
inline typename std::iterator_traits<RanIt>::value_type fat_approx_med_(RanIt begin, RanIt mid, RanIt end, Pred comp)
{	// pseudomedian of thirty-one
	insertion_sort(begin,begin+11,comp);
	insertion_sort(mid-5,mid+5,comp);
	insertion_sort(end-11,end,comp);
	med3_(begin+5, mid, end-6, comp);
	return *mid;
}

template<class RanIt, class Pred>
inline typename std::iterator_traits<RanIt>::value_type approx_med_(RanIt begin, RanIt end, Pred comp)
{
	RanIt mid = begin+(end-begin)/2;
	if( end-begin < 96 )
		return median_(*begin, *mid, *(end-1), comp);
	else if( end-begin < 2048 )
	{	// pseudomedian of nine
		int step = (end-begin)/8;
		--end;
		med3_(begin, begin+step, begin+2*step, comp);
		med3_(mid-step, mid, mid+step, comp);
		med3_(end-2*step, end-step, end, comp);
		med3_(begin+step, mid, end-step, comp);
	}
	else
		return fat_approx_med_(begin,mid,end,comp);
	return *mid;
}

template<class BidIt, class T, class Pred>
std::pair<BidIt,BidIt> fat_partition_(BidIt begin, BidIt end, T pivot, Pred comp)
{
	BidIt first, last, eqend, eqbegin;
	eqend = first = begin;
	eqbegin = last = end;
	for( ;; )
	{
		assert( first <= last );
		for( ; first<last && !comp(pivot,*first); ++first )
			if( !comp(*first,pivot) )
				std::iter_swap(eqend,first), ++eqend;
		assert( last==first || (comp(pivot,*first) && first<last) );
		for( --last; first<last && !comp(*last,pivot); --last )
			if( !comp(pivot,*last) )
				std::iter_swap(--eqbegin,last);
		assert( last<=first || (comp(*last,pivot) && first<last) );
		if( last <= first )
			break;
		else
		{
			assert( comp(*last,pivot) && comp(pivot,*first) );
			std::iter_swap(first, last);
			assert( comp(*first,pivot) && comp(pivot,*last) );
			++first;
		}
	}
/*	BidIt i;
	for( i=begin; i!=eqend; ++i )
		assert( *i == pivot );
	for( ; i!=first; ++i )
		assert( *i < pivot );
	for( ; i!=eqbegin; ++i )
		assert( pivot < *i );
	for( ; i!=end; ++i )
		assert( *i == pivot );
*/	std::pair<BidIt,BidIt> ret = std::make_pair(first,first);
	if( eqend != begin )
	{
		for( --eqend, --ret.first; eqend!=begin; --eqend, --ret.first )
			std::iter_swap(eqend,ret.first);
		std::iter_swap(eqend,ret.first);
	}
	for( ; eqbegin!=end; ++eqbegin, ++ret.second )
		std::iter_swap(ret.second,eqbegin);
/*	for( i=begin; i!=ret.first; ++i )
		assert( *i < pivot );
	for( ; i!=ret.second; ++i )
		assert( *i == pivot );
	for( ; i!=end; ++i )
		assert( pivot < *i );
*/	return ret;
}

template<class BidIt, class T, class Pred>
inline std::pair<BidIt,BidIt> partition_(BidIt begin, BidIt end, T pivot, Pred comp)
{
	if( end-begin < 2048 )
		for( ;; )
		{
			for( ; comp(*begin,pivot); ++begin );
			for( --end; comp(pivot,*end); --end );
			if( !(begin<end) )
				return std::make_pair(begin,begin);
			std::iter_swap(begin, end);
			++begin;
		}
	else
		return fat_partition_(begin,end,pivot,comp);
}


template<class RanIt, class Diff, class Pred>
inline void intro_sort_(RanIt First, RanIt Last, Diff Ideal, Pred comp)
{
	Diff Count;
	while( (Count = Last-First) > RECURSELIMIT_ && Ideal > 0 )
	{
		std::pair<RanIt,RanIt> part = partition_(First, Last, approx_med_(First,Last,comp),comp);
		Ideal /= 2, Ideal += Ideal / 2; // allow 1.5 log2(N) divisions
		if( part.first-First < Last-part.second ) // loop on larger half
			intro_sort_<RanIt,Diff>(First, part.first, Ideal, comp), First = part.second;
		else
			intro_sort_<RanIt,Diff>(part.second, Last, Ideal, comp), Last = part.first;
	}
	if( Count > RECURSELIMIT_ )
	{
		std::make_heap(First, Last, comp);
		std::sort_heap(First, Last, comp);
	}
	else if( Count > 1 )
        insertion_sort(First, Last, comp);
}

template<class RanIt>
inline void intro_sort(RanIt begin, RanIt end)
{
	intro_sort_(begin,end,end-begin,std::less<typename std::iterator_traits<RanIt>::value_type>());
/*	typename std::iterator_traits<RanIt>::value_type prev = *begin;
	for( ++begin; begin!=end; prev=*begin, ++begin )
		assert(!( *begin<prev ));
*/}

#include "../bench.h"

struct sorter
{
	typedef minxp::true_type inplace;
	template<class RanIt>
	static inline void sort(RanIt b, RanIt e)
		{ intro_sort(b, e); }
	static const char* name;
};
const char* sorter::name = "introsort";

int main(int argc, char *argv[])
{
	return bench_this<sorter>(argv,argc+argv);
}
