
/*
  * fast in-memory sorting
  * (C) 1999 Rajiv Wickremesinghe
  * Duke University
  */

/* $Id: merge2.c,v 1.9 2000/04/07 01:54:31 rajiv Exp $ */

#include <minxp/array.h>
#include "../bench.h"
#include <iterator>
#include <utility>

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <assert.h>
#include <math.h>
#include <string.h>
//#include "common.h"
//#include "const.h"
//#include "memory.h"
#include "heap.h"



//
//   !!! FIXES !!!
//

#define MEASURE_DECLARE(what)
#define MEASURE_START(what)
#define MEASURE_END(what)
#define MEASURE_ADD(what)
#define ALLOCATE_MEMORY(where, howmuch) where = new ELT[howmuch/sizeof(ELT)]
int expd(int e, int d)
{
	int r = 1;
	for( ; d; --d, r*=e );
	return r;
}
template<class ELT>
int is_sorted(ELT *arr, int n, void *b)
{
	for( int i=1; i<n; ++i )
		if( arr[i] < arr[i-1] )
			return 0;
	return 1;
}

#include <limits>
#include <iterator>
#define SENTINEL std::numeric_limits<ELT>::max()

enum tag_RECURSELIMIT_ { M = 24 };

template<class It, class Pred>
inline void insertion_sort_(It begin, It end, Pred comp)
{
	for( It j=begin+1; j<end; j++ )
	{
		typename std::iterator_traits<It>::value_type e = *j;
		for( It i=j; begin<i; *i=e )
			if( comp(e,*(i-1)) )
				*i = *(i-1), --i;
			else
				break;
	}
}

template<class ELT>
inline void insertion_sort(ELT *a, int n)
{
	insertion_sort_(a, a+n, std::less<ELT>());
}


//
//   !!! END FIXES !!!
//


MEASURE_DECLARE(run);
MEASURE_DECLARE(memload);

#ifdef NDEBUG
#define DEBUG if(1)
#else
#define DEBUG
#endif

typedef struct {
    int l,r;
    } StackElt;

#define Push(nl,nr)  {stackPtr++;stack[stackPtr].l=nl;stack[stackPtr].r=nr;}
#define Pop(nl,nr) {nl=stack[stackPtr].l;nr=stack[stackPtr].r;stackPtr--;}
#define Empty (stackPtr<0)
#define swap(x,y) {ELT tt; tt=a[x];a[x]=a[y];a[y]=tt;}


template<class ELT>
void compactQuicksort(ELT *a, int N) {
   int l,r,i,j/*,k*/;
   /*int done;*/
   ELT /*t,*/v;
   StackElt stack[1000];
   int      stackPtr=-1;
   /*int gl,gr;*/
   /*ELT tmp;*/

   Push(0,N-1);
   while (!Empty) {

     Pop(l,r);

     if ((r-l) <= M) {
       /* small list, lets sort in now while we've got it in the cache */
	  insertion_sort(a+l, r-l+1);
       /* its all sorted now. We're done with this segment */
       continue;
     }

     /* choose the pivot using median of three */
	{
	  swap((l+r)/2,l+1);
	  if (a[r] < a[l+1]) swap(l+1,r);
	  if (a[r] < a[l]) swap(l,r);
	  if (a[l] < a[l+1]) swap(l+1,l);
	  i = l+1;
	  j = r;
	  v = a[l];
	}

     do i++; while (a[i] < v);
	do j--; while (v < a[j]);
	while(i < j) {
       swap(i,j);
       do i++; while (a[i] < v);
       do j--; while (v < a[j]);
     }
     swap(l,j);

     /* do big side first */
     if (j-l < r-i+1) {
       Push(i-1,r);
       Push(l,j); /* right side is smaller, do left first */
     } else {
       Push(l,j);
       Push(i-1,r); /* left side is smaller, do right first */
     }
   }
   /* insertion sort not nexxesary now, list is sorted */
   assert(is_sorted(a, N, NULL));
}



/* 
********************************************************************** 
*/

/* #define ACCOUNTING */

#ifdef ACCOUNTING
enum { POS_0=0, POS_1, POS_2, POS_3, POS_4, POS_5, POS_6, POS_7,
	   POS_new, POS_last };
static int count[POS_last][POS_last];
#define ACCOUNT(a,b) (count[POS_##a][POS_##b]++)

void print_account() {
   int i,j;
   for(i=POS_0; i<POS_last; i++) {
	printf(" %d ", i);
	for(j=i+1; j<POS_last; j++) {
	  if(count[j][i]) printf(" %d:%4.2f", j, (double)count[j][i]/(1<<20));
	}
	printf("\n");
   }
}

#else
#define ACCOUNT(a,b) 1
#endif

/* 
********************************************************************** 
*/




/* 
********************************************************************** 
*/


	
/* merge data from the heapsize_g streams in a[] into outarr[]
    the total number of elements, N,  should be greater than order.
    size is the length of each stream
  */
template<class ELT>
void merge7(ELT *a, int N, ELT* outarr, int size) {
#ifndef NDEBUG
   int dummycnt=0;
   ELT *t, *d, prevmin = std::numeric_limits<ELT>::min();
#endif
   ELT dnew;
   int snew;
   DEFINEHEAP
   int i;
   /*ELT tmp;*/
   ELT save;

   assert(N>heapsize_g);
   assert(N==size*heapsize_g);



   /* build the heap */
#ifndef NDEBUG
//   for(i=0; i<N; i++) assert(0<a[i]);
   ALLOCATE_MEMORY(t,sizeof(ELT) * N);
   memcpy(t, a, N*sizeof(ELT));
   ALLOCATE_MEMORY(d,sizeof(ELT) * heapsize_g);
#endif
   save = a[heapsize_g*size];
   BUILDHEAP_OPT();
#ifdef PRINT_VERBOSE
   for(i=1; i<N+1; i++) {
	if(i%size == 0) {
	  assert(a[i] == SENTINEL);
	} else {
	  assert(a[i] != SENTINEL);
	  if(i==N || a[i] == t[i]); else {
		printf("i=%d\n", i);
		for(i=0; i<30; i++) {
		  printf("%d %8d %8d\n", i, a[i], t[i]);
		}
		assert(i==N || a[i] == t[i]);
	  }
	}
   }

	HEAP2ARRAY(d);
   d[0] = dnew;

   {
	int j;
	printf("input:\n");
	for(i=0; i<N; i+= size) {
	  printf("%d: ", i);
	  for(j=0; j<size; j++) {
		printf("%d ", t[i+j]);
	  }
	  printf("\n");
	}
	printf("heap: ");
	for(i=0; i<heapsize_g; i++) {
	  printf("%d ", d[i]);
	}
	printf("\n");
   }
   tmp = 0;
   for(i=0; i<heapsize_g; i++) {
	tmp += t[i*size] - d[i];
   }
   if(tmp) {
	for(i=0; i<heapsize_g; i++) {
	  printf("%ld %ld\n", t[i*size], d[i]);
	}
	assert(tmp==0);
   }

   fflush(stdout);
#endif

   for(i=0; i<N; i++) {
	HEAPIFY_OPT();
	*outarr++ = MIN_VAL();
	dnew = a[MIN_SRC()];
#ifndef NDEBUG
//	assert(0 < MIN_VAL());
	if(!(MIN_VAL() < prevmin)); else {
	  printf("i=%d; MIN_SRC=%d\n", i, MIN_SRC());
	  assert(!(MIN_VAL() < prevmin));
	}
	prevmin = MIN_VAL();
	if(MIN_VAL() == SENTINEL) {
	  printf("i=%d, N=%d\n", i, N);
	  assert(!(MIN_VAL() == SENTINEL));
	}
	assert(!(SENTINEL < dnew));
//	assert(0 < dnew);
	if(dnew == SENTINEL) {
	  assert(MIN_SRC()%size == 0);
	  dummycnt++;
	  assert(dummycnt <= heapsize_g);
	}
#endif
	snew = MIN_SRC() + 1;
   }
   a[heapsize_g*size] = save;
}


template<class ELT>
void
merge7sort_helper(ELT *arr1, int N, int runsize, int depth, ELT 
*outarray) {
   int i/*,j*/;
   int size7, size49;
   int arity = heapsize_g;

   assert(runsize * expd(arity, depth) == N);
   assert((depth%2) == 0);
   assert((N%expd(arity, depth)) == 0);

#ifndef NDEBUG
//   for(i=0; i<N; i++) assert(0<arr1[i]);
#endif

   /* Start of run formation */
   MEASURE_START(run);
   for(i=0; i<N; i+=runsize) {
#ifndef NDEBUG
	for(int j=0; j<N; j++) {
//	  assert(0 < arr1[j]);
	}
#endif
	compactQuicksort(arr1 + i, runsize);
	assert(is_sorted(arr1+i, runsize, NULL));
#ifndef NDEBUG
	for(int j=0; j<N; j++) {
//	  assert(0 < arr1[j]);
	}
#endif
   }
   MEASURE_END(run);
   MEASURE_ADD(run);
   /* End of run formation */


#ifdef MERGE_ONLY
    StartUsage();
    StartCounting();
#endif

	size7 = runsize*arity;
	size49 = runsize;
	for( ; depth; depth-=2 ) {
		for(i=0; i<N; i+=size7) {
#ifndef NDEBUG
			printf("pre merge at i=%d\n", i);
			for(int j=0; j<N; j++) {
//				assert(0 < arr1[j]);
			}
#endif
			merge7(arr1+i, size7, outarray+i, size49);
			assert(is_sorted(outarray+i, size7, NULL));
#ifndef NDEBUG
			for(int j=0; j<N; j++) {
//				assert(0 < arr1[j]);
			}
#endif
		}
		size7 *= arity;
		size49 *= arity;

		for(i=0; i<N; i+=size7) {
			merge7(outarray+i, size7, arr1+i, size49);
			assert(is_sorted(arr1+i, size7, NULL));
		}
		size7 *= arity;
		size49 *= arity;
	}
#ifdef MERGE_ONLY
   EndCounting();
   EndUsage();
#endif

#ifdef ACCOUNTING
   print_account();
#endif

   assert(is_sorted(arr1, N, NULL));
}

/*
void
merge7sort(ELT *arr1, int N, int runsize, int depth) {
   ELT *outarray;

   ALLOCATE_MEMORY(outarray, N*sizeof(ELT));
   merge7sort_helper(arr1, N, runsize, depth, outarray);
   * DEALLOCATE_MEMORY *
}
*/

template<class ELT>
void merge7_ord(int order, ELT *a, int N, ELT* outarr, int size);

template<class ELT>
void merge7_ord(int order, ELT *a, int N, ELT* outarr, int size) {
   switch(order) {
   case 3:
	merge_3(a, N, outarr, size);
	break;
   case 4:
	merge_4(a, N, outarr, size);
	break;
   case 5:
	merge_5(a, N, outarr, size);
	break;
   case 7:
	merge_7(a, N, outarr, size);
	break;
   case 9:
	merge_9(a, N, outarr, size);
	break;
   case 11:
	merge_11(a, N, outarr, size);
	break;
   case 13:
	merge_13(a, N, outarr, size);
	break;
   case 15:
	merge_15(a, N, outarr, size);
	break;
   case 17:
	merge_17(a, N, outarr, size);
	break;
   case 19:
	merge_19(a, N, outarr, size);
	break;
   case 21:
	merge_21(a, N, outarr, size);
	break;
   default:
	fprintf(stderr, "unsupported merge order: %d\n", order);
	exit(1);
   }
}

template<class ELT>
void
merge7sort_2pass(ELT *arr1, int N, int paddedN,
				 int runsize1, int order1,
				 int runsize2, int order2) {
   int i,j;
   ELT *outarray;

   ALLOCATE_MEMORY(outarray, paddedN*sizeof(ELT)+1);

   /* Start of run formation */
   MEASURE_START(run);
   for(i=0; i<N; i+=runsize1) {
	compactQuicksort(arr1 + i, runsize1);
	assert(is_sorted(arr1+i, runsize1, NULL));
   }
   MEASURE_END(run);
   MEASURE_ADD(run);
   /* End of run formation */


   /* first pass */
   for(i=0; i<N; i+=runsize2) {
	merge7_ord(order1, arr1+i, runsize2, outarray+i, runsize1);
	assert(is_sorted(outarray+i, runsize2, NULL));
   }
   /* second pass */
   merge7_ord(order2, outarray, N, arr1, runsize2);

   assert(is_sorted(arr1, N, NULL));
}





template<class ELT>
void
merge7sort_single(ELT *arr1, int N) {
   int i,j;
   ELT *outarray;
   int runsize;
   int arity = heapsize_g;

   runsize = N/arity;
   assert(N==arity*runsize);
   ALLOCATE_MEMORY(outarray, N*sizeof(ELT));

   /* Start of run formation */
   MEASURE_START(run);
   for(i=0; i<N; i+=runsize) {
	compactQuicksort(arr1+i, runsize);
	assert(is_sorted(arr1+i, runsize, NULL));
   }
   MEASURE_END(run);
   MEASURE_ADD(run);
   /* End of run formation */


#ifdef MERGE_ONLY
#error foo
	StartUsage();
	StartCounting();
#endif
	
	merge7(arr1, N, outarray, runsize);
	assert(is_sorted(outarray, N, NULL));

#ifdef MERGE_ONLY
   EndCounting();
   EndUsage();
#endif

   /* memcpy(arr1, outarray, N*sizeof(ELT)); */

}

template<class ELT>
void
variable_arity_merge(int arity, ELT *inarr, int N, ELT *outarr, int 
runlength)
{
   switch(arity) {
   case 1:
	memcpy(outarr, inarr, N * sizeof(ELT));
	break;
   case 3:
	merge_3(inarr, N, outarr, runlength);
	break;
   case 5:
	merge_5(inarr, N, outarr, runlength);
	break;
   case 7:
	merge_7(inarr, N, outarr, runlength);
	break;
   case 9:
	merge_9(inarr, N, outarr, runlength);
	break;
   case 11:
	merge_11(inarr, N, outarr, runlength);
	break;
   default:
	assert(0);
	break;
   }
}

template<class ELT>
void
network_sort(ELT *arr1, int N, int nruns) {
   int i;
   int mm_runsize, cache_runsize;
   int depth=2;
   ELT *inp, *outp, *arr2;
   int arity=heapsize_g;

   ALLOCATE_MEMORY(arr2, N*sizeof(ELT));
   assert(arr2);

   /* form runs of main memory size */
   mm_runsize = N/nruns;
   cache_runsize = mm_runsize/(arity*arity);

   DEBUG printf("nruns=%d, mm_runsize=%.2fM\n", nruns,
			   (double)mm_runsize/(1<<20));
   DEBUG printf("cache_runsize=%5.3fM (%.2fMB)\n",
			   (double)cache_runsize/(1<<20),
			   (double)cache_runsize*sizeof(ELT)/(1<<20));

   /* sort runs using best internal memory mergesort */
   MEASURE_START(memload);
   for(i=0, inp=arr1, outp=arr2;
	  i<nruns;
	  i++, inp+=mm_runsize, outp+=mm_runsize) {
	DEBUG printf("sorting memoryload #%d\n", i);
	merge7sort_helper(inp, mm_runsize, cache_runsize, depth, outp);
	assert(is_sorted(inp, mm_runsize, NULL));
   }
   MEASURE_END(memload);

   /* merge the runs */
   DEBUG printf("doing final merge\n");
   variable_arity_merge(nruns, arr1, N, arr2, mm_runsize);
   /* memcpy(arr1, arr2, N*sizeof(ELT)); */
}

struct old_sorter
{
	typedef minxp::false_type inplace;
	template<class RanIt, class OutIt>
	static inline void sort(RanIt b, RanIt e, OutIt out)
	{
		merge7sort_helper(b, e-b, runsize<RanIt>(), depth<RanIt,7>(e-b), out);
	}
	static const char* name;
private:
	static const int cache_size = 256*1024;
	template<class It>
	inline static int runsize()
		{ return cache_size/sizeof(typename std::iterator_traits<It>::value_type); };
	template<class It, int arity>
	inline static int depth(int N)
	{
		assert( N % runsize<It>() == 0 );
		N /= runsize<It>();
		int d;
		for( d=0; N > 1; ++d, N/=arity );
		return d;
	}
};
const char* old_sorter::name = "rmerge";
struct sorter
{
	typedef minxp::false_type inplace;
	typedef std::random_access_iterator_tag out_iterator_category;
	template<class RanIt, class OutIt>
	static inline void sort(RanIt b, RanIt e, OutIt out)
	{
		std::pair<typename std::iterator_traits<RanIt>::difference_type,typename std::iterator_traits<RanIt>::difference_type> sizes = runsize<2,7>(e-b);
		if( sizes.second > cachesize<RanIt>() )
		{
			sizes = runsize<4,7>(e-b);
			merge7sort_helper(b, static_cast<int>(sizes.first), static_cast<int>(sizes.second), 4, out);
		}
		else
		{
			merge7sort_helper(b, static_cast<int>(sizes.first), static_cast<int>(sizes.second), 2, out);
		}
	}
	static const char* name;
private:
	static const int cache_size = 256*1024;
	template<class It>
	inline static int cachesize()
		{ return cache_size/sizeof(typename std::iterator_traits<It>::value_type); };
	template<int k, int arity, class ItDiff>
	inline static std::pair<ItDiff,ItDiff> runsize(ItDiff d)
	{
		std::pair<ItDiff,ItDiff> ret;
		for( int i=0; i!=k; ++i )
			d /= arity;
		ret.second = d;
		for( int i=0; i!=k; ++i )
			d *= arity;
		ret.first = d;
		return ret;
	};
};
const char* sorter::name = "rmerge";

int main(int argc, char *argv[])
{
	return bench_this<sorter>(argv,argc+argv);
}
