/*
 * This code is from "Algorithms in C++, Third Edition,"
 * by Robert Sedgewick, Addison-Wesley, 1999.
 *
 * available at
 * http://www.cs.princeton.edu/~rs/Algs3.cxx1-4/code.txt
 *
 * Insignificant changes by Kristoffer Vinther, 2003
 */

#include <iterator>

template <class Item> 
  void exch(Item &A, Item &B) 
    { Item t = A; A = B; B = t; } 
template <class Item> 
  void compexch(Item &A, Item &B) 
    { if (B < A) exch(A, B); }

template <class It>
void insertion(It a, int l, int r)
  { typedef typename std::iterator_traits<It>::value_type Item;
    int i;
    for (i = r; i > l; i--) compexch(a[i-1], a[i]);
    for (i = l+2; i <= r; i++)
      { int j = i; Item v = a[i]; 
        while (v < a[j-1])
          { a[j] = a[j-1]; j--; }
        a[j] = v; 
      } 
  }

template <class It>
int partition(It a, int l, int r)
  { int i = l-1, j = r; typename std::iterator_traits<It>::value_type v = a[r];
    for (;;)
      { 
        while (a[++i] < v) ;
        while (v < a[--j]) if (j == l) break;
        if (i >= j) break;
        exch(a[i], a[j]);
      }
    exch(a[i], a[r]);
    return i;
  }

static const int M = 10;
template <class It>
void quicksort(It a, int l, int r)
  {
    if (r-l <= M) return;
    exch(a[(l+r)/2], a[r-1]);
    compexch(a[l], a[r-1]); 
      compexch(a[l], a[r]); 
        compexch(a[r-1], a[r]);
    int i = partition(a, l+1, r-1);
    quicksort(a, l, i-1);
    quicksort(a, i+1, r);
  } 
template <class It>
void hybridsort(It a, int l, int r)
  { quicksort(a, l, r); insertion(a, l, r); }


#include "minxp/array.h"
#include "../bench.h"

struct sorter
{
	typedef minxp::true_type inplace;
	template<class RanIt>
	static inline void sort(RanIt b, RanIt e)
		{ hybridsort(b, 0, static_cast<int>(e-b)); }
	static const char* name;
};
const char* sorter::name = "sedsort";

int main(int argc, char *argv[])
{
	return bench_this<sorter>(argv,argc+argv);
}
