#ifndef __SORTED_STREAM_H_INCLUDED_
#define __SORTED_STREAM_H_INCLUDED_

#include <iterator>

template<class T>
class sorted_stream
{
public:
	class iterator : public std::iterator<std::random_access_iterator_tag, T, int>
	{
		friend class sorted_stream<T>;
	public:
		inline iterator& operator++()
		{
			if( cnt )
			{
				if( s->prev > s->current )
					std::cerr << "Elements " << static_cast<int>(cnt) << " and " << static_cast<int>(cnt+1) << " out of order: " << s->prev << '>' << s->current << std::endl;
				s->prev = s->current;
				s->chksum += s->current;
			}
			else
			{
				s->prev = s->current;
				s->chksum = s->current;
			}
			++cnt;
			return *this;
		}
		inline iterator operator+(int i) const
		{
			iterator it = *this;
			it.cnt += i;
			return it;
		}
		inline T& operator*()
			{ return s->current; }
		inline const T& operator*() const
			{ return s->current; }
		inline bool operator==(const iterator& rhs) const
			{ return cnt == rhs.cnt; }
		inline bool operator!=(const iterator& rhs) const
			{ return cnt != rhs.cnt; }
		inline long operator-(const iterator& rhs) const
			{ return static_cast<long>(cnt-rhs.cnt); }
	protected:
		iterator(sorted_stream *s) : s(s), cnt(0) { }
		iterator(sorted_stream *s, size_t c) : s(s), cnt(c) { }
	private:
		sorted_stream *s;
		size_t cnt;
	};

	sorted_stream(size_t s) : size(s) { }
	iterator begin()
		{ return iterator(this); }
	iterator end()
		{ return iterator(this, size); }
	inline T checksum() const
		{ return chksum; }
protected:
	T prev;
	T current;
	T chksum;
	size_t size;
};

#endif // __SORTED_STREAM_H_INCLUDED_
